__author__ = 'Dmitriy Korsakov'

import os

import yaml

from scalrctl import click
from scalrctl import defaults
from scalrctl import settings
from scalrctl import commands
from scalrctl.commands.internal import bash_complete, update


class ConfigureScalrCTL(commands.BaseAction):

    def run(self, *args, **kwargs):
        configure(**kwargs)

    def get_description(self):
        return "Set configuration options in interactive mode"

    def get_options(self):
        profile_argument = click.Argument(("profile",), required=False)  # [ST-30]
        return [profile_argument, ]


def configure(profile=None):
    """
    Configure command-line client.
    Creates new profile in configuration directory
    and downloads spec file.
    :param profile: Profile name
    """
    confpath = os.path.join(defaults.CONFIG_DIRECTORY, "%s.yaml" % profile) if profile else defaults.CONFIG_PATH
    data = {}

    if os.path.exists(confpath):
        old_data = yaml.load(open(confpath, "r"))
        data.update(old_data)

    click.echo("Configuring %s:" % confpath)
    excludes = ("SSL_VERIFY_PEER", "SIGNATURE_VERSION", "API_VERSION", "debug_mode")
    for obj in dir(settings):
        if not obj.startswith("__") and obj not in excludes:
            default_value = getattr(settings, obj)
            if obj == "view":
                click.echo("Set your default view mode. Choose from tree, table, json")
                view = str(click.prompt("view", default="tree"))
                if view not in ("tree", "table", "json"):
                    view = "tree"
                data["view"] = view
            elif isinstance(default_value, bool):
                data[obj] = click.confirm(obj, default=getattr(settings, obj))
            elif not default_value or type(default_value) in (int, str):
                data[obj] = str(click.prompt(obj, default=getattr(settings, obj))).strip()
            if obj == "API_SCHEME" and data[obj] == "https":
                data["SSL_VERIFY_PEER"] = click.confirm("SSL_VERIFY_PEER", default=True)

    if not os.path.exists(defaults.CONFIG_DIRECTORY):
        os.makedirs(defaults.CONFIG_DIRECTORY)

    raw = yaml.dump(data, default_flow_style=False, default_style='')
    with open(confpath, 'w') as fp:
        fp.write(raw)

    click.echo()
    click.echo("New config saved:")
    click.echo()
    click.echo(open(confpath, "r").read())

    apply_settings(data)

    update.update()
    bash_complete.setup_bash_complete()


def apply_settings(data):
    for key, value in data.items():
        if hasattr(settings, key):
            setattr(settings, key, value)



__author__ = 'Dmitriy Korsakov'

import os
import json
import pydoc

import yaml

from scalrctl import click
from scalrctl import settings
from scalrctl import defaults
from scalrctl import commands

from scalrctl.commands.internal import update, configure


CMD_FOLDER = os.path.abspath(os.path.join(os.path.dirname(__file__), 'commands'))


if not os.path.exists(defaults.CONFIG_DIRECTORY):
    os.makedirs(defaults.CONFIG_DIRECTORY)

if os.path.exists(defaults.CONFIG_PATH):
    configure.apply_settings(yaml.load(open(defaults.CONFIG_PATH, "r")))

if update.is_update_required():
    update.update()  # [ST-53]


def dummy_run():
    raise click.ClickException("Not implemented in current API version")

class ScalrCLI(click.Group):

    def __init__(self, name=None, commands=None, **attrs):
        if not "scheme" in attrs:
            with open(os.path.join(os.path.dirname(__file__), 'scheme/scheme.json')) as fp:
                self.scheme = json.load(fp)
        else:
            self.scheme = attrs.pop("scheme")
        super(ScalrCLI, self).__init__(name, commands, **attrs)

    def list_commands(self, ctx):
        commands = self.scheme.keys()

        if "group_descr" in commands:
            commands.remove("group_descr")

        if "route" in commands and "http-method" in commands:
            return []

        commands.sort()
        return commands

    def get_command(self, ctx, name):
        args = dict(
            callback=lambda: None,
        )

        if name not in self.scheme:

            if ctx.protected_args:
                for level in ctx.protected_args:
                    if level in self.scheme:
                        self.scheme = self.scheme[level]
                    else:
                        ctx.exit()

        if name not in self.scheme:
            click.echo("No such command: %s." % name)
            ctx.exit()

        args["scheme"] = self.scheme[name]

        if "group_descr" in self.scheme[name]:
            args["short_help"] = self.scheme[name]["group_descr"]

        action_level = "route" in self.scheme[name] and "http-method" in self.scheme[name]
        is_service_type_action = action_level and not self.scheme[name]["api_level"]

        if action_level:
            if is_service_type_action:
                route = None
                http_method = None
                api_level = None
            else:
                route = self.scheme[name]["route"]
                http_method = self.scheme[name]["http-method"]
                api_level = self.scheme[name]["api_level"]

            if 'class' in self.scheme[name]:
                cls = pydoc.locate(self.scheme[name]['class'])
            else:
                cls = commands.Action

            action = cls(name=name, route=route, http_method=http_method, api_level=api_level)
            hidden = "hidden" in self.scheme[name] and self.scheme[name]['hidden']

            try:
                action.validate()
            except AssertionError as e:
                dummy_help = "Not implemented in current API version"
                dummy_cmd = click.Command(name, params=[], callback=dummy_run, short_help=dummy_help, hidden=hidden)
                return dummy_cmd

            hlp = self.scheme[name]["cmd_descr"] if "cmd_descr" in self.scheme[name] else action.get_description()
            options = action.modify_options(action.get_options())

            cmd = click.Command(name, params=options, callback=action.run, short_help=hlp, hidden=hidden)
            if action.epilog:
                cmd.epilog = action.epilog
            return cmd

        return ScalrCLI(**args)


@click.command(cls=ScalrCLI)
@click.version_option()
@click.pass_context
@click.option('--key_id', help="API key ID")
@click.option('--secret_key', help="API secret key")
@click.option('--config', help="Path to a custom scalr-ctl configuration file")
def cli(ctx, key_id, secret_key, config, *args, **kvargs):
    """Scalr-ctl is a command-line interface to your Scalr account"""

    if key_id:
        settings.API_KEY_ID = str(key_id)

    if secret_key:
        settings.API_SECRET_KEY = str(secret_key)

    elif settings.API_KEY_ID and settings.API_KEY_ID.strip() and not settings.API_SECRET_KEY:  # [ST-21]
        if ctx.invoked_subcommand not in ("configure", "update"):
            raw = click.prompt(text="API SECRET KEY", hide_input=True)
            settings.API_SECRET_KEY = str(raw)

    if config:

        if not os.path.exists(config):
            raise click.ClickException("Configuration file not found: %s" % config)
        data = yaml.load(open(config, "r"))
        configure.apply_settings(data)


if __name__ == '__main__':
    cli()



__author__ = 'Dmitriy Korsakov'

import json
import datetime
import copy

from scalrctl import settings
from scalrctl import commands
from scalrctl import click
from scalrctl import defaults

import yaml


class Export(commands.Action):

    def _get_custom_options(self):
        # Disable output modifiers
        options = []
        debug = click.Option(('--debug/--no-debug', 'debug'), default=False, help="Print debug messages")
        options.append(debug)
        return options

    def run(self, *args, **kwargs):
        hide_output = kwargs.pop("hide_output", False)

        kv = kwargs.copy()
        kv['hide_output'] = True
        response = super(Export, self).run(*args, **kv)

        kwargs["envId"] = settings.envId
        kv = copy.deepcopy(kwargs)
        for item in ("debug", "nocolor", "transformation"):
            if item in kv:
                del kv[item]
        uri = self._request_template.format(**kwargs)

        d = {
            "API_VERSION": settings.API_VERSION,
            "envId": settings.envId,
            "DATE": datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S"),
            "API_HOST": settings.API_HOST,
            "API_LEVEL": self.api_level,
            "METHOD": self.http_method,
            "ROUTE": self.route,
            "URI": uri,
            "ACTION": self.name,
            "ARGUMENTS": (args, kv),
            "SCALRCTL_VERSION": defaults.VERSION,
             }

        try:
            response_json = json.loads(response)
        except ValueError as e:
            if settings.debug_mode:
                raise
            raise click.ClickException(str(e))

        response_json["meta"]["scalrctl"] = d

        dump = yaml.safe_dump(
            response_json,
            encoding='utf-8',
            allow_unicode=True,
            default_flow_style=False
        )
        if not hide_output:
            click.echo(dump)
        return response_json


class ExportFarmRoleGlobalVariable(Export):
    prompt_for = ["roleId", "globalVariableName"]


class ExportImage(Export):
    prompt_for = ["imageId"]


class ExportScript(Export):

    def run(self, *args, **kwargs):
        # TODO: help for update! (some script-versions may be deleted)
        # interactive update by default (prompt for deletes and updates), with optional force
        kwargs["hide_output"] = True
        response = super(ExportScript, self).run(*args, **kwargs)

        svlist_action = commands.Action(
            "script-version",
            "/{envId}/scripts/{scriptId}/script-versions/",
            "get",
            "user"
        )
        svlist_text = svlist_action.run(envId=kwargs["envId"], scriptId=kwargs["scriptId"], hide_output=True)
        svlist_dict = json.loads(svlist_text)
        svlist = svlist_dict["data"]

        svget_action = Export(
            "script-version",
            "/{envId}/scripts/{scriptId}/script-versions/{scriptVersionNumber}/",
            "get",
            "user"
        )
        include = []
        for sv in svlist:
            script_version = sv["version"]
            sctipt_body = sv["body"]
            if sctipt_body:
                out = svget_action.run(
                    envId=kwargs["envId"],
                    scriptId=kwargs["scriptId"],
                    scriptVersionNumber=script_version,
                    hide_output=True
                )
                include.append(out)

        response["include"] = include

        dump = yaml.safe_dump(
            response,
            encoding='utf-8',
            allow_unicode=True,
            default_flow_style=False
        )
        click.echo(dump)
        return response



# -*- coding: utf-8 -*-
# This module is based on the excellent work by Adam Bartoš who
# provided a lot of what went into the implementation here in
# the discussion to issue1602 in the Python bug tracker.
#
# There are some general differences in regards to how this works
# compared to the original patches as we do not need to patch
# the entire interpreter but just work in our little world of
# echo and prmopt.

import io
import os
import sys
import zlib
import time
import ctypes
import msvcrt
from ._compat import _NonClosingTextIOWrapper, text_type, PY2
from ctypes import byref, POINTER, c_int, c_char, c_char_p, \
     c_void_p, py_object, c_ssize_t, c_ulong, windll, WINFUNCTYPE
try:
    from ctypes import pythonapi
    PyObject_GetBuffer = pythonapi.PyObject_GetBuffer
    PyBuffer_Release = pythonapi.PyBuffer_Release
except ImportError:
    pythonapi = None
from ctypes.wintypes import LPWSTR, LPCWSTR


c_ssize_p = POINTER(c_ssize_t)

kernel32 = windll.kernel32
GetStdHandle = kernel32.GetStdHandle
ReadConsoleW = kernel32.ReadConsoleW
WriteConsoleW = kernel32.WriteConsoleW
GetLastError = kernel32.GetLastError
GetCommandLineW = WINFUNCTYPE(LPWSTR)(
    ('GetCommandLineW', windll.kernel32))
CommandLineToArgvW = WINFUNCTYPE(
    POINTER(LPWSTR), LPCWSTR, POINTER(c_int))(
        ('CommandLineToArgvW', windll.shell32))


STDIN_HANDLE = GetStdHandle(-10)
STDOUT_HANDLE = GetStdHandle(-11)
STDERR_HANDLE = GetStdHandle(-12)


PyBUF_SIMPLE = 0
PyBUF_WRITABLE = 1

ERROR_SUCCESS = 0
ERROR_NOT_ENOUGH_MEMORY = 8
ERROR_OPERATION_ABORTED = 995

STDIN_FILENO = 0
STDOUT_FILENO = 1
STDERR_FILENO = 2

EOF = b'\x1a'
MAX_BYTES_WRITTEN = 32767


class Py_buffer(ctypes.Structure):
    _fields_ = [
        ('buf', c_void_p),
        ('obj', py_object),
        ('len', c_ssize_t),
        ('itemsize', c_ssize_t),
        ('readonly', c_int),
        ('ndim', c_int),
        ('format', c_char_p),
        ('shape', c_ssize_p),
        ('strides', c_ssize_p),
        ('suboffsets', c_ssize_p),
        ('internal', c_void_p)
    ]

    if PY2:
        _fields_.insert(-1, ('smalltable', c_ssize_t * 2))


# On PyPy we cannot get buffers so our ability to operate here is
# serverly limited.
if pythonapi is None:
    get_buffer = None
else:
    def get_buffer(obj, writable=False):
        buf = Py_buffer()
        flags = PyBUF_WRITABLE if writable else PyBUF_SIMPLE
        PyObject_GetBuffer(py_object(obj), byref(buf), flags)
        try:
            buffer_type = c_char * buf.len
            return buffer_type.from_address(buf.buf)
        finally:
            PyBuffer_Release(byref(buf))


class _WindowsConsoleRawIOBase(io.RawIOBase):

    def __init__(self, handle):
        self.handle = handle

    def isatty(self):
        io.RawIOBase.isatty(self)
        return True


class _WindowsConsoleReader(_WindowsConsoleRawIOBase):

    def readable(self):
        return True

    def readinto(self, b):
        bytes_to_be_read = len(b)
        if not bytes_to_be_read:
            return 0
        elif bytes_to_be_read % 2:
            raise ValueError('cannot read odd number of bytes from '
                             'UTF-16-LE encoded console')

        buffer = get_buffer(b, writable=True)
        code_units_to_be_read = bytes_to_be_read // 2
        code_units_read = c_ulong()

        rv = ReadConsoleW(self.handle, buffer, code_units_to_be_read,
                          byref(code_units_read), None)
        if GetLastError() == ERROR_OPERATION_ABORTED:
            # wait for KeyboardInterrupt
            time.sleep(0.1)
        if not rv:
            raise OSError('Windows error: %s' % GetLastError())

        if buffer[0] == EOF:
            return 0
        return 2 * code_units_read.value


class _WindowsConsoleWriter(_WindowsConsoleRawIOBase):

    def writable(self):
        return True

    @staticmethod
    def _get_error_message(errno):
        if errno == ERROR_SUCCESS:
            return 'ERROR_SUCCESS'
        elif errno == ERROR_NOT_ENOUGH_MEMORY:
            return 'ERROR_NOT_ENOUGH_MEMORY'
        return 'Windows error %s' % errno

    def write(self, b):
        bytes_to_be_written = len(b)
        buf = get_buffer(b)
        code_units_to_be_written = min(bytes_to_be_written,
                                       MAX_BYTES_WRITTEN) // 2
        code_units_written = c_ulong()

        WriteConsoleW(self.handle, buf, code_units_to_be_written,
                      byref(code_units_written), None)
        bytes_written = 2 * code_units_written.value

        if bytes_written == 0 and bytes_to_be_written > 0:
            raise OSError(self._get_error_message(GetLastError()))
        return bytes_written


class ConsoleStream(object):

    def __init__(self, text_stream, byte_stream):
        self._text_stream = text_stream
        self.buffer = byte_stream

    @property
    def name(self):
        return self.buffer.name

    def write(self, x):
        if isinstance(x, text_type):
            return self._text_stream.write(x)
        try:
            self.flush()
        except Exception:
            pass
        return self.buffer.write(x)

    def writelines(self, lines):
        for line in lines:
            self.write(line)

    def __getattr__(self, name):
        return getattr(self._text_stream, name)

    def isatty(self):
        return self.buffer.isatty()

    def __repr__(self):
        return '<ConsoleStream name=%r encoding=%r>' % (
            self.name,
            self.encoding,
        )


def _get_text_stdin(buffer_stream):
    text_stream = _NonClosingTextIOWrapper(
        io.BufferedReader(_WindowsConsoleReader(STDIN_HANDLE)),
        'utf-16-le', 'strict', line_buffering=True)
    return ConsoleStream(text_stream, buffer_stream)


def _get_text_stdout(buffer_stream):
    text_stream = _NonClosingTextIOWrapper(
        _WindowsConsoleWriter(STDOUT_HANDLE),
        'utf-16-le', 'strict', line_buffering=True)
    return ConsoleStream(text_stream, buffer_stream)


def _get_text_stderr(buffer_stream):
    text_stream = _NonClosingTextIOWrapper(
        _WindowsConsoleWriter(STDERR_HANDLE),
        'utf-16-le', 'strict', line_buffering=True)
    return ConsoleStream(text_stream, buffer_stream)


if PY2:
    def _hash_py_argv():
        return zlib.crc32('\x00'.join(sys.argv[1:]))

    _initial_argv_hash = _hash_py_argv()

    def _get_windows_argv():
        argc = c_int(0)
        argv_unicode = CommandLineToArgvW(GetCommandLineW(), byref(argc))
        argv = [argv_unicode[i] for i in range(0, argc.value)]

        if not hasattr(sys, 'frozen'):
            argv = argv[1:]
            while len(argv) > 0:
                arg = argv[0]
                if not arg.startswith('-') or arg == '-':
                    break
                argv = argv[1:]
                if arg.startswith(('-c', '-m')):
                    break

        return argv[1:]


_stream_factories = {
    0: _get_text_stdin,
    1: _get_text_stdout,
    2: _get_text_stderr,
}


def _get_windows_console_stream(f, encoding, errors):
    if get_buffer is not None and \
       encoding in ('utf-16-le', None) \
       and errors in ('strict', None) and \
       hasattr(f, 'isatty') and f.isatty():
        func = _stream_factories.get(f.fileno())
        if func is not None:
            if not PY2:
                f = getattr(f, 'buffer')
                if f is None:
                    return None
            else:
                # If we are on Python 2 we need to set the stream that we
                # deal with to binary mode as otherwise the exercise if a
                # bit moot.  The same problems apply as for
                # get_binary_stdin and friends from _compat.
                msvcrt.setmode(f.fileno(), os.O_BINARY)
            return func(f)
